#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <err.h>

#define DEG2RAD(x) (x*M_PI/180.0)

void
check_magnitude(const char *func_name, int limit, int value)
{
	if (value < limit)
		errx(1, "%s: input size %d is less than %d\n",
		     func_name, value, limit);
}

/* Translate a i,j,k index in grid with dimensions nx, ny, nz into a
 * linear index */
unsigned int
idx3(const unsigned int i,
     const unsigned int j,
     const unsigned int k,
     const unsigned int nx,
     const unsigned int ny)
{
	return i + nx * j + nx * ny * k;
}

/* Translate a i,j,k index in grid with dimensions nx, ny, nz and a
 * padding of single ghost nodes into a linear index */
unsigned int
idx3g(const unsigned int i,
      const unsigned int j,
      const unsigned int k,
      const unsigned int nx,
      const unsigned int ny)
{
	return i + 1 + (nx + 2) * (j + 1) + (nx + 2) * (ny + 2) * (k + 1);
}

/* Translate a i,j index in grid with dimensions nx, ny into a linear
 * index */
unsigned int
idx2(const unsigned int i, const unsigned int j, const unsigned int nx)
{
	return i + nx * j;
}

/* Translate a i,j index in grid with dimensions nx, ny and a padding of
 * single ghost nodes into a linear index */
unsigned int
idx2g(const unsigned int i, const unsigned int j, const unsigned int nx)
{
	return i + 1 + (nx + 2) * (j + 1);
}

/* Translate a i index in grid with a padding of single into a linear
 * index */
unsigned int
idx1g(const unsigned int i)
{
	return i + 1;
}

/* Return an array of `n` linearly spaced values in the range [lower;
 * upper] */
double *
linspace(const double lower, const double upper, const int n)
{
	int i;
	double *x;
	double dx;

	check_magnitude(__func__, 1, n);
	if (!(x = calloc(n, sizeof(double))))
		err(1, "%s: x calloc", __func__);
	dx = (upper - lower) / (double) (n - 1);
	for (i = 0; i < n; ++i)
		x[i] = lower + dx * i;

	return x;
}

/* Return an array of `n-1` values with the intervals between `x` values */
double *
spacing(const double *x, const int n)
{
	int i;
	double *dx;

	check_magnitude(__func__, 2, n);
	if (!(dx = calloc((n - 1), sizeof(double))))
		err(1, "%s: dx calloc", __func__);
	for (i = 0; i < n - 1; ++i)
		dx[i] = x[i + 1] - x[i];

	return dx;
}

/* Return an array of `n` values with the value 0.0 */
double *
zeros(const int n)
{
	int i;
	double *x;

	check_magnitude(__func__, 1, n);
	if (!(x = calloc(n, sizeof(double))))
		err(1, "%s: x calloc", __func__);
	for (i = 0; i < n; ++i)
		x[i] = 0.0;

	return x;
}

/* Return an array of `n` values with the value 1.0 */
double *
ones(const int n)
{
	int i;
	double *x;

	check_magnitude(__func__, 1, n);
	if (!(x = calloc(n, sizeof(double))))
		err(1, "%s: x calloc", __func__);
	for (i = 0; i < n; ++i)
		x[i] = 1.0;

	return x;
}

/* Return an array of `n` values with a specified value */
double *
initval(const double value, const int n)
{
	int i;
	double *x;

	check_magnitude(__func__, 1, n);
	if (!(x = calloc(n, sizeof(double))))
		err(1, "%s: x calloc", __func__);
	for (i = 0; i < n; ++i)
		x[i] = value;

	return x;
}

/* Return an array of `n` uninitialized values */
double *
empty(const int n)
{
	double *out;

	check_magnitude(__func__, 1, n);

	if (!(out = calloc(n, sizeof(double))))
		err(1, "%s: calloc", __func__);

	return out;
}

/* Return largest value in array `a` with size `n` */
double
max(const double *a, const int n)
{
	int i;
	double maxval;

	check_magnitude(__func__, 1, n);
	maxval = -INFINITY;
	for (i = 0; i < n; ++i)
		if (a[i] > maxval)
			maxval = a[i];

	return maxval;
}

/* Return smallest value in array `a` with size `n` */
double
min(const double *a, const int n)
{
	int i;
	double minval;

	check_magnitude(__func__, 1, n);
	minval = +INFINITY;
	for (i = 0; i < n; ++i)
		if (a[i] < minval)
			minval = a[i];

	return minval;
}

void
print_array(const double *a, const int n)
{
	int i;

	check_magnitude(__func__, 1, n);
	for (i = 0; i < n; ++i)
		printf("%.17g\n", a[i]);
}

void
print_arrays(const double *a, const double *b, const int n)
{
	int i;

	check_magnitude(__func__, 1, n);
	for (i = 0; i < n; ++i)
		printf("%.17g\t%.17g\n", a[i], b[i]);
}

void
print_arrays_2nd_normalized(const double *a, const double *b, const int n)
{
	int i;
	double max_b;

	check_magnitude(__func__, 1, n);
	max_b = max(b, n);
	for (i = 0; i < n; ++i)
		printf("%.17g\t%.17g\n", a[i], b[i] / max_b);
}

void
print_three_arrays(const double *a,
                   const double *b,
                   const double *c,
                   const int n)
{
	int i;

	check_magnitude(__func__, 1, n);
	for (i = 0; i < n; ++i)
		printf("%.17g\t%.17g\t%.17g\n", a[i], b[i], c[i]);
}

void
fprint_arrays(FILE *fp, const double *a, const double *b, const int n)
{
	int i;

	check_magnitude(__func__, 1, n);
	for (i = 0; i < n; ++i)
		fprintf(fp, "%.17g\t%.17g\n", a[i], b[i]);
}

void
fprint_three_arrays(FILE *fp,
                    const double *a,
                    const double *b,
                    const double *c,
                    const int n)
{
	int i;

	check_magnitude(__func__, 1, n);
	for (i = 0; i < n; ++i)
		fprintf(fp, "%.17g\t%.17g\t%.17g\n", a[i], b[i], c[i]);
}

void
copy_values(const double *in, double *out, const int n)
{
	int i;

	check_magnitude(__func__, 1, n);
	for (i = 0; i < n; ++i)
		out[i] = in[i];
}

double *
copy(const double *in, const int n)
{
	double *out;

	check_magnitude(__func__, 1, n);
	out = empty(n);
	copy_values(in, out, n);
	return out;
}

double *
normalize(const double *in, const int n)
{
	int i;
	double max_val;
	double *out;

	check_magnitude(__func__, 1, n);
	out = empty(n);
	copy_values(in, out, n);
	max_val = max(out, n);

	if (max_val == 0.0)
		max_val = 1.0;

	for (i = 0; i < n; ++i)
		out[i] /= max_val;

	return out;
}

double
euclidean_norm(const double *a, const int n)
{
	int i;
	double out = 0.0;

	for (i = 0; i < n; i++)
		out += pow(a[i], 2.0);

	return sqrt(out);
}

double
euclidean_distance(const double *a, const double *b, const int n)
{
	int i;
	double out = 0.0;

	for (i = 0; i < n; i++)
		out += pow(b[i] - a[i], 2.0);

	return sqrt(out);
}

double
dot(const double *a, const double *b, const int n)
{
	int i;
	double out = 0.0;

	for (i = 0; i < n; i++)
		out += a[i] * b[i];

	return out;
}

double *
cross(const double a[3], const double b[3])
{
	double *out = empty(3);

	out[0] = a[1]*b[2] - a[2]*b[1];
	out[1] = a[2]*b[0] - a[0]*b[2];
	out[2] = a[0]*b[1] - a[1]*b[0];

	return out;
}
