#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <err.h>
#include "granular.h"
#include "util.h"
#include "grain.h"
#include "arrays.h"
#include "contact.h"

#define FLOATPREC 17

struct grain
grain_new(void)
{
	struct grain g;
	grain_defaults(&g);
	return g;
}

void
grain_defaults(struct grain *g)
{
	int d;

	g->diameter = 1.0;
	for (d = 0; d < 3; d++) {
		g->pos[d]
		= g->vel[d]
		= g->acc[d]
		= g->force[d]
		= g->angpos[d]
		= g->angvel[d]
		= g->angacc[d]
		= g->torque[d]
		= g->disp[d]
		= g->forceext[d]
		= g->contact_stress[d]
		= 0.0;
		g->gridpos[d]
		= g->acc_lock[d] = 0;
	}
	g->density = 2.5e3;
	g->fixed = 0;
	g->rotating = 1;
	g->enabled = 1;
	g->youngs_modulus = 1e9;
	g->poissons_ratio = 0.2;
	g->friction_coeff = 0.4;
	g->damping_n = 0.0;
	g->damping_t = 0.0;
	g->tensile_strength = 1e8;
	g->shear_strength = 1e8;
	g->fracture_toughness = 1e8;
	g->ncontacts = 0;
	for (d = 0; d < MAXCONTACTS; d++)
		contact_defaults(&g->contacts[d]);
	g->thermal_energy = 0.0;
	g->color = 0;
}

static void
print_padded_nd_double(FILE *stream, const double *arr)
{
	int d;
	
	for (d = 0; d < 3; d++)
		fprintf(stream, "%.*g\t", FLOATPREC, arr[d]);
}

static void
print_padded_nd_int(FILE *stream, const int *arr)
{
	int d;
	
	for (d = 0; d < 3; d++)
		fprintf(stream, "%d\t", arr[d]);
}

static void
print_padded_nd_uint(FILE *stream, const size_t *arr)
{
	int d;
	
	for (d = 0; d < 3; d++)
		fprintf(stream, "%zu\t", arr[d]);
}

void
grain_print(FILE *stream, const struct grain *g)
{
	fprintf(stream, "%.*g\t", FLOATPREC, g->diameter);
	print_padded_nd_double(stream, g->pos);
	print_padded_nd_double(stream, g->vel);
	print_padded_nd_double(stream, g->acc);
	print_padded_nd_int(stream, g->acc_lock);
	print_padded_nd_double(stream, g->force);
	print_padded_nd_double(stream, g->angpos);
	print_padded_nd_double(stream, g->angvel);
	print_padded_nd_double(stream, g->angacc);
	print_padded_nd_double(stream, g->torque);
	print_padded_nd_double(stream, g->disp);
	print_padded_nd_double(stream, g->forceext);
	fprintf(stream, "%.*g\t", FLOATPREC, g->density);
	fprintf(stream, "%d\t", g->fixed);
	fprintf(stream, "%d\t", g->rotating);
	fprintf(stream, "%d\t", g->enabled);
	fprintf(stream, "%.*g\t", FLOATPREC, g->youngs_modulus);
	fprintf(stream, "%.*g\t", FLOATPREC, g->poissons_ratio);
	fprintf(stream, "%.*g\t", FLOATPREC, g->friction_coeff);
	fprintf(stream, "%.*g\t", FLOATPREC, g->damping_n);
	fprintf(stream, "%.*g\t", FLOATPREC, g->damping_t);
	fprintf(stream, "%.*g\t", FLOATPREC, g->tensile_strength);
	fprintf(stream, "%.*g\t", FLOATPREC, g->shear_strength);
	fprintf(stream, "%.*g\t", FLOATPREC, g->fracture_toughness);
	print_padded_nd_uint(stream, g->gridpos);
	fprintf(stream, "%zu\t", g->ncontacts);
	print_padded_nd_double(stream, g->contact_stress);
	fprintf(stream, "%.*g\t", FLOATPREC, g->thermal_energy);
	fprintf(stream, "%d\n", g->color);
}

struct grain *
grain_read(char *line)
{
	struct grain *g;

	if (!(g = malloc(sizeof(struct grain))))
		err(1, "%s: grain malloc", __func__);

	if (sscanf(line, "%lg\t"           /* diameter */
	                 "%lg\t%lg\t%lg\t" /* pos */
	                 "%lg\t%lg\t%lg\t" /* vel */
	                 "%lg\t%lg\t%lg\t" /* acc */
	                 "%d\t%d\t%d\t"    /* acc_lock */
	                 "%lg\t%lg\t%lg\t" /* force */
	                 "%lg\t%lg\t%lg\t" /* angpos */
	                 "%lg\t%lg\t%lg\t" /* angvel */
	                 "%lg\t%lg\t%lg\t" /* angacc */
	                 "%lg\t%lg\t%lg\t" /* torque */
	                 "%lg\t%lg\t%lg\t" /* disp */
	                 "%lg\t%lg\t%lg\t" /* forceext */
	                 "%lg\t"           /* density */
	                 "%d\t"            /* fixed */
	                 "%d\t"            /* rotating */
	                 "%d\t"            /* enabled */
	                 "%lg\t"           /* youngs_modulus */
	                 "%lg\t"           /* poissons_ratio */
	                 "%lg\t"           /* friction_coeff */
	                 "%lg\t"           /* damping_n */
	                 "%lg\t"           /* damping_t */
	                 "%lg\t"           /* tensile_strength */
	                 "%lg\t"           /* shear_strength */
	                 "%lg\t"           /* fracture_toughness */
	                 "%zu\t%zu\t%zu\t" /* gridpos */
	                 "%zu\t"           /* ncontacts */
	                 "%lg\t%lg\t%lg\t" /* contact_stress */
	                 "%lg\t"           /* thermal_energy */
	                 "%d\n",           /* color */
	                 &g->diameter,
	                 &g->pos[0], &g->pos[1], &g->pos[2], 
	                 &g->vel[0], &g->vel[1], &g->vel[2], 
	                 &g->acc[0], &g->acc[1], &g->acc[2], 
	                 &g->acc_lock[0], &g->acc_lock[1], &g->acc_lock[2], 
	                 &g->force[0], &g->force[1], &g->force[2], 
	                 &g->angpos[0], &g->angpos[1], &g->angpos[2], 
	                 &g->angvel[0], &g->angvel[1], &g->angvel[2], 
	                 &g->angacc[0], &g->angacc[1], &g->angacc[2], 
	                 &g->torque[0], &g->torque[1], &g->torque[2], 
	                 &g->disp[0], &g->disp[1], &g->disp[2], 
	                 &g->forceext[0], &g->forceext[1], &g->forceext[2], 
	                 &g->density,
	                 &g->fixed,
	                 &g->rotating,
	                 &g->enabled,
	                 &g->youngs_modulus,
	                 &g->poissons_ratio,
	                 &g->friction_coeff,
	                 &g->damping_n,
	                 &g->damping_t,
	                 &g->tensile_strength,
	                 &g->shear_strength,
	                 &g->fracture_toughness,
	                 &g->gridpos[0], &g->gridpos[1], &g->gridpos[2], 
	                 &g->ncontacts,
	                 &g->contact_stress[0], &g->contact_stress[1], &g->contact_stress[2], 
	                 &g->thermal_energy,
	                 &g->color) != 55)
		errx(1, "%s: could not read line: %s", __func__, line);

	if (grain_check_values(g))
		errx(1, "%s: invalid values in grain line: %s", __func__, line);

	return g;
}

int
grain_check_values(const struct grain *g)
{
	int d;
	int status = 0;

	check_float_positive("grain->diameter", g->diameter, &status);

	for (d = 0; d < 3; d++) {
		check_float("grain->pos", g->pos[d], &status);
		check_float("grain->vel", g->vel[d], &status);
		check_float("grain->acc", g->acc[d], &status);
		check_int_bool("grain->acc_lock", g->acc_lock[d], &status);
		check_float("grain->force", g->force[d], &status);
		check_float("grain->angpos", g->angpos[d], &status);
		check_float("grain->angvel", g->angvel[d], &status);
		check_float("grain->angacc", g->angacc[d], &status);
		check_float("grain->torque", g->torque[d], &status);
		check_float("grain->disp", g->disp[d], &status);
		check_float("grain->forceext", g->forceext[d], &status);
		check_float("grain->contact_stress", g->contact_stress[d], &status);
	}

	check_float_non_negative("grain->density", g->density, &status);
	check_int_bool("grain->fixed", g->fixed, &status);
	check_int_bool("grain->rotating", g->rotating, &status);
	check_int_bool("grain->enabled", g->enabled, &status);

	check_float_non_negative("grain->youngs_modulus",
	                         g->youngs_modulus,
	                         &status);
	check_float_non_negative("grain->poissons_ratio",
	                         g->poissons_ratio,
	                         &status);
	check_float_non_negative("grain->friction_coeff",
	                         g->friction_coeff,
	                         &status);
	check_float_non_negative("grain->damping_n",
	                         g->damping_n,
	                         &status);
	check_float_non_negative("grain->damping_t",
	                         g->damping_t,
	                         &status);
	check_float_non_negative("grain->tensile_strength",
	                         g->tensile_strength,
	                         &status);
	check_float_non_negative("grain->shear_strength",
	                         g->shear_strength,
	                         &status);
	check_float_non_negative("grain->fracture_toughness",
	                         g->fracture_toughness,
	                         &status);

	for (d = 0; d < 3; d++)
		if (g->gridpos[d] > 1)
			warn_parameter_value("grain->gridpos is not 0 or 1",
								 (double)g->gridpos[d], &status);

	check_float_non_negative("grain->thermal_energy", g->thermal_energy, &status);

	return status;
}

double
grain_volume(const struct grain *g)
{
	return 4.0 / 3.0 * PI * pow(g->diameter / 2.0, 3.0);
}

double
grain_mass(const struct grain *g)
{
	return grain_volume(g) * g->density;
}

double
grain_moment_of_inertia(const struct grain *g)
{
	return 2.0 / 5.0 * grain_mass(g) * pow(g->diameter / 2.0, 2.0);
}

void
grain_zero_kinematics(struct grain *g)
{
	int d;

	for (d = 0; d < 3; d++) {
		g->vel[d] = 0.0;
		g->acc[d] = 0.0;
		g->force[d] = 0.0;
		g->angvel[d] = 0.0;
		g->angacc[d] = 0.0;
		g->torque[d] = 0.0;
		g->disp[d] = 0.0;
	}
}

void
grain_force_reset(struct grain *g)
{
	int d;

	for (d = 0; d < 3; d++)
		g->force[d] = 0.0;
}

double
grain_translational_kinetic_energy(const struct grain *g)
{
	return 0.5 * grain_mass(g) * euclidean_norm(g->vel, 3.0);
}

double
grain_rotational_kinetic_energy(const struct grain *g)
{
	return 0.5 * grain_moment_of_inertia(g) * euclidean_norm(g->angvel, 3.0);
}

double
grain_kinetic_energy(const struct grain *g)
{
	return grain_translational_kinetic_energy(g) +
		grain_rotational_kinetic_energy(g);
}

double
grain_thermal_energy(const struct grain *g)
{
	return g->thermal_energy;
}

static void
grain_temporal_integration_two_term_taylor(struct grain *g, double dt)
{
	int d;
	double dx, mass = grain_mass(g);
	double moment_of_inertia = grain_moment_of_inertia(g);

	for (d = 0; d < 3; d++) {
		g->acc[d] = (g->force[d] + g->forceext[d]) / mass;

		if (g->rotating)
			g->angacc[d] = g->torque[d] / moment_of_inertia;
	}

	if (g->fixed)
		for (d = 0; d < 3; d++) {
			g->angacc[d] = 0.0;
			if (!g->acc_lock[d])
				g->acc[d] = 0.0;
		}

	for (d = 0; d < 3; d++) {
		dx = g->vel[d] * dt + 0.5 * g->acc[d] * dt * dt;
		g->pos[d] += dx;
		g->disp[d] += dx;
		g->angpos[d] += g->angvel[d] * dt + 0.5 * g->angacc[d] * dt * dt;
		g->vel[d] += g->acc[d] * dt;
		g->angvel[d] += g->angacc[d] * dt;
	}
}

void
grain_temporal_integration(struct grain *g, double dt)
{
	grain_temporal_integration_two_term_taylor(g, dt);
}

void
grain_register_contact(struct grain *g, size_t i, size_t j,
                       double centerdist[3], double overlap)
{
	int ic, d;

	/* first pass: check if contact is already registered */
	for (ic = 0; ic < MAXCONTACTS; ic++)
		if (g->contacts[ic].active && g->contacts[ic].j == j) {
			g->contacts[ic].overlap = overlap;
			for (d = 0; d < 3; d++)
				g->contacts[ic].centerdist[d] = centerdist[d];
			return;
		}

	/* second pass: register as new contact if overlapping */
	if (overlap > 0.0)
		for (ic = 0; ic <= MAXCONTACTS; ic++) {
			if (ic == MAXCONTACTS)
				err(1, "%s: contact %zu exceeds MAXCONTACTS (%d)",
				    __func__, i, MAXCONTACTS);
			if (!g->contacts[ic].active) {
				contact_new(&g->contacts[ic], i, j);
				g->ncontacts += 1;
				g->contacts[ic].overlap = overlap;
				for (d = 0; d < 3; d++) {
					g->contacts[ic].centerdist[d] = centerdist[d];
					g->contacts[ic].tandisp[d] = 0.0;
				}
				return;
			}
		}
}

double
grain_stiffness_normal(const struct grain *g)
{
	return 2.0 / 3.0 * g->youngs_modulus / (1.0 - pow(g->poissons_ratio, 2.0));
}

double
grain_stiffness_tangential(const struct grain *g)
{
	return 2.0 * g->youngs_modulus
	       / ((1.0 + g->poissons_ratio) * (2.0 - g->poissons_ratio));
}

double
grain_collision_time(const struct grain *g)
{
	double m;

	m = grain_mass(g);

	return fmin(PI * pow(2.0 * grain_stiffness_normal(g) / m
				         - pow(g->damping_n, 2.0) / 4.0, -0.5),
				PI * pow(2.0 * grain_stiffness_tangential(g) / m
				         - pow(g->damping_t, 2.0) / 4.0, -0.5));
}
