#ifndef __OpenBSD__
#define _POSIX_C_SOURCE 200809L
#endif
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <err.h>
#include <string.h>
#include <math.h>
#include "grain.h"
#include "grains.h"
#include "simulation.h"
#include "arrays.h"
#include "util.h"

struct simulation
sim_new(void)
{
	struct simulation sim;
	sim_defaults(&sim);
	return sim;
}

void
sim_defaults(struct simulation *sim)
{
	int d, ret;

	ret = snprintf(sim->name, sizeof(sim->name), DEFAULT_SIMULATION_NAME);
	if (ret < 0 || (size_t)ret >= sizeof(sim->name))
		errx(1, "%s: could not write simulation name", __func__);
	sim->ng = 0;
	for (d = 0; d < 3; d++) {
		sim->nd[d] = 1;
		sim->origo[d] = 0.0;
		sim->L[d] = 1.0;
	}
	sim->t = 0.0;
	sim->t_end = 0.0;
	sim->dt = 0.0;
	sim->dt_file = 1.0;
	sim->n_file = 0;
	sim->iter = 0;
	sim->t_file = 0;
	sim->ng = 0;
	sim->grains = NULL;
}

void
sim_free(struct simulation *sim)
{
	free(sim->grains);
	sim->grains = NULL;
	sim->ng = 0;
}

void
sim_add_grain(struct simulation *sim, struct grain *g)
{
	if (!(sim->grains = xreallocarray(sim->grains, sim->ng + 1, sizeof(*g))))
		err(1, "%s: sim.grains reallocarray", __func__);
	memcpy(&sim->grains[sim->ng++], g, sizeof(*g));
}

void
sim_read_grains(struct simulation *sim, FILE *stream)
{
	char *line = NULL;
	size_t linesize = 0;
	ssize_t linelen;
	struct grain *g;

	while ((linelen = getline(&line, &linesize, stream)) > 0) {
		g = grain_read(line);
		sim_add_grain(sim, g);
		free(g);
	}

	free(line);
}

void
sim_print_grains(const struct simulation *sim, FILE *stream)
{
	grains_print(stream, sim->grains, sim->ng);
}

void
sim_print_grains_vtk(const struct simulation *sim, FILE *stream)
{
	grains_print_vtk(stream, sim->grains, sim->ng);
}

void
sim_print_grains_energy(const struct simulation *sim, FILE *stream)
{
	grains_print_energy(stream, sim->grains, sim->ng);
}

void
sim_write_output_files(struct simulation *sim)
{
	int ret;
	char outfile[256];
	FILE *fp;

	ret = snprintf(outfile, sizeof(outfile), "%s.grains.%05zu.tsv",
	               sim->name, sim->n_file++);
	if (ret < 0 || (size_t)ret >= sizeof(outfile))
		errx(1, "%s: could not write grains filename", __func__);
	if ((fp = fopen(outfile, "w")) != NULL) {
		sim_print_grains(sim, fp);
		fclose(fp);
	} else
		err(1, "%s: fopen: %s", __func__, outfile);
}

double
sim_grain_pair_overlap(struct simulation *sim, size_t i, size_t j)
{
	double pos[3], overlap;
	int d;

	for (d = 0; d < 3; d++)
		pos[d] = sim->grains[i].pos[d] - sim->grains[j].pos[d];
	overlap = (sim->grains[i].diameter + sim->grains[j].diameter) / 2.0
			- euclidean_norm(pos, 3);

	return overlap;
}

static void
sim_check_add_contact(struct simulation *sim, size_t i, size_t j)
{

	double overlap, centerdist[3];
	int d;

	if (i >= j || !sim->grains[i].enabled || !sim->grains[j].enabled)
		return;

	for (d = 0; d < 3; d++)
		centerdist[d] = sim->grains[i].pos[d] - sim->grains[j].pos[d];

	overlap = 0.5*(sim->grains[i].diameter + sim->grains[j].diameter)
	        - euclidean_norm(centerdist, 3);

	grain_register_contact(&sim->grains[i], i, j, centerdist, overlap);
}

void
sim_force_reset(struct simulation *sim)
{
	size_t i;
	for (i = 0; i < sim->ng; i++)
		grain_force_reset(&sim->grains[i]);
}

void
sim_add_acceleration_scalar(struct simulation *sim, double acc, int dimension)
{
	size_t i;

	for (i = 0; i < sim->ng; i++)
		sim->grains[i].forceext[dimension]
			+= acc * grain_mass(&sim->grains[i]);
}

/* Silbert et al. 2001 */
void
sim_set_timestep(struct simulation *sim)
{
	int ret = 0;
	size_t i;
	double t_col, t_col_min = INFINITY;

	for (i = 0; i < sim->ng; i++) {
		t_col = grain_collision_time(&sim->grains[i]);
		if (t_col_min > t_col)
			t_col_min = t_col;
	}

	check_float("t_col_min", t_col_min, &ret);
	sim->dt = t_col_min / 50.0;
}

/* TODO: add grid sorting */
void
sim_detect_contacts(struct simulation *sim)
{
	size_t i, j;

	for (i = 0; i < sim->ng; i++)
		if (sim->grains[i].enabled)
			for (j = 0; j < sim->ng; j++)
				if (i < j)
					sim_check_add_contact(sim, i, j);
}

void
sim_resolve_interactions(struct simulation *sim)
{
	size_t i;
	int ic;

	for (i = 0; i < sim->ng; i++)
		for (ic = 0; ic < MAXCONTACTS; ic++)
			if (sim->grains[i].contacts[ic].active && i < sim->grains[i].contacts[ic].j)
				grains_interact(&sim->grains[i],
				                &sim->grains[sim->grains[i].contacts[ic].j],
				                ic, sim->dt);
}

void
sim_step_time(struct simulation *sim)
{
	size_t i;

	sim_force_reset(sim);
	sim_detect_contacts(sim);
	sim_resolve_interactions(sim);

	for (i = 0; i < sim->ng; i++)
		grain_temporal_integration(&sim->grains[i], sim->dt);

	if ((sim->t_file >= sim->dt_file || sim->iter == 0) &&
		strncmp(sim->name, DEFAULT_SIMULATION_NAME,
				sizeof(sim->name)) != 0) {
		sim_write_output_files(sim);
		sim->t_file = 0.0;
	}

	sim->t += sim->dt;
	sim->t_file += sim->dt;
	sim->iter++;
}

void
sim_run_time_loop(struct simulation *sim)
{
	if (sim->dt > sim->dt_file)
		sim->dt = sim->dt_file;

	do {
		sim_step_time(sim);
	} while (sim->t - sim->dt < sim->t_end);
}
